import logging
from datetime import datetime
from datetime import timezone

from pymyenergi.connection import Connection

from . import CT_BATTERY
from . import CT_GENERATION
from . import CT_GRID
from . import CT_LOAD
from . import DEVICE_TYPES
from . import EDDI
from . import FREQUENCY_GRID
from . import HARVI
from . import LIBBI
from . import HOUR
from . import VOLTAGE_GRID
from . import ZAPPI
from .eddi import Eddi
from .harvi import Harvi
from .libbi import Libbi
from .zappi import Zappi

_LOGGER = logging.getLogger(__name__)


def device_factory(conn, kind, serial, data=None):
    """Create device instances"""
    if kind == ZAPPI:
        return Zappi(conn, serial, data)
    if kind == EDDI:
        return Eddi(conn, serial, data)
    if kind == HARVI:
        return Harvi(conn, serial, data)
    if kind == LIBBI:
        return Libbi(conn, serial, data)
    raise Exception(f"Unsupported device type {kind}")


class MyenergiClient:
    """Zappi Client for myenergi API."""

    def __init__(
        self,
        connection: Connection,
    ) -> None:
        self._connection = connection
        self.devices = {}
        self._data = []
        self._keys = None
        self._totals = {}
        self._history_totals = {}
        self._update_available = False
        self._firmware_version = ""

    @property
    def site_name(self):
        """myenergi API site name"""
        return self.find_device_name("siteName", f"Hub_{self._connection.username}")

    @property
    def serial_number(self):
        """Hub serial number"""
        return self._connection.username

    @property
    def firmware_version(self):
        """Hub firmware version"""
        return self._firmware_version

    def _calculate_history_totals(self):
        """Caluclate history data totals"""
        devices = self.get_devices_sync()
        energy_keys = ["generated", "grid_import", "grid_export"]
        self._history_totals = {"green": 0}
        for key in energy_keys:
            self._history_totals[key] = 0
        zappi_or_eddi_or_libbi = None
        for device in devices:
            if device.kind == ZAPPI or device.kind == EDDI or device == LIBBI:
                zappi_or_eddi_or_libbi = device
                break
        if zappi_or_eddi_or_libbi is not None:
            for key in energy_keys:
                if self._history_totals[key] == 0:
                    self._history_totals[key] = zappi_or_eddi_or_libbi.history_data.get(key, 0)
            self._history_totals["green"] = self._history_totals.get(
                "green", 0
            ) + zappi_or_eddi_or_libbi.history_data.get("device_green", 0)

    def _calculate_totals(self):
        """Calculate current data totals"""
        devices = self.get_devices_sync()
        self._totals = {}
        zappi_or_eddi_or_libbi = None
        self._totals[CT_GRID] = 0
        self._totals[CT_GENERATION] = 0
        for device in devices:
            if device.ct1.is_assigned:
                self._totals[device.ct1.name] = (
                    self._totals.get(device.ct1.name, 0) + device.ct1.power
                )
            if device.ct2.is_assigned:
                self._totals[device.ct2.name] = (
                    self._totals.get(device.ct2.name, 0) + device.ct2.power
                )

            if device.kind in [ZAPPI, HARVI, LIBBI]:
                if device.ct3.is_assigned:
                    self._totals[device.ct3.name] = (
                        self._totals.get(device.ct3.name, 0) + device.ct3.power
                    )
            if device.kind == EDDI:
                zappi_or_eddi_or_libbi = device
            if device.kind == ZAPPI or device.kind == LIBBI:
                zappi_or_eddi_or_libbi = device
                if device.kind == ZAPPI or device.kind == LIBBI:
                    if device.ct4.is_assigned:
                        self._totals[device.ct4.name] = (
                            self._totals.get(device.ct4.name, 0) + device.ct4.power
                        )
                    if device.ct5.is_assigned:
                        self._totals[device.ct5.name] = (
                            self._totals.get(device.ct5.name, 0) + device.ct5.power
                        )
                    if device.ct6.is_assigned:
                        self._totals[device.ct6.name] = (
                            self._totals.get(device.ct6.name, 0) + device.ct6.power
                        )

        if zappi_or_eddi_or_libbi is not None:
            self._totals[FREQUENCY_GRID] = zappi_or_eddi_or_libbi.supply_frequency
            self._totals[VOLTAGE_GRID] = zappi_or_eddi_or_libbi.supply_voltage
        if self._totals.get(CT_GRID, 0) == 0 and zappi_or_eddi_or_libbi is not None:
            self._totals[CT_GRID] = zappi_or_eddi_or_libbi.power_grid
        if self._totals.get(CT_GENERATION, 0) == 0 and zappi_or_eddi_or_libbi is not None:
            self._totals[CT_GENERATION] = zappi_or_eddi_or_libbi.power_generated

    def get_power_totals(self):
        return self._totals

    @property
    def consumption_home(self):
        """Calculates home power"""
        # calculation is all generation + grid + battery - device consumption
        return (
            self._totals.get(CT_GENERATION, 0)
            + self._totals.get(CT_GRID, 0)
            + self._totals.get(CT_BATTERY, 0)
            - self._totals.get(CT_LOAD, 0)
        )

    @property
    def energy_imported(self):
        """Grid imported energy"""
        return self._history_totals.get("grid_import", 0)

    @property
    def energy_exported(self):
        """Grid exported energy"""
        return self._history_totals.get("grid_export", 0)

    @property
    def energy_generated(self):
        """Generated energy"""
        return self._history_totals.get("generated", 0)

    @property
    def energy_green(self):
        """Green energy"""
        return self._history_totals.get("green", 0)

    @property
    def power_grid(self):
        """Grid total power"""
        return self._totals.get(CT_GRID, 0)

    @property
    def frequency_grid(self):
        """Grid frequency"""
        return self._totals.get(FREQUENCY_GRID, None)

    @property
    def voltage_grid(self):
        """Grid frequency"""
        return self._totals.get(VOLTAGE_GRID, None)

    @property
    def power_generation(self):
        """Generation total power"""
        return self._totals.get(CT_GENERATION, 0)

    @property
    def power_charging(self):
        """Chargers total power"""
        return self._totals.get(CT_LOAD, 0)

    @property
    def power_battery(self):
        """Battery total power"""
        return self._totals.get(CT_BATTERY, 0)
    

    def find_device_name(self, key, default_value):
        """Find device or site name"""
        keys = list(self._keys.values())[0]
        return next((item["val"] for item in keys if item["key"] == key), default_value)

    async def refresh(self):
        """Refresh device data"""
        _LOGGER.debug("Refreshing data for all myenergi devices")
        data = await self.fetch_data()
        self._data = data["devices"]
        self._keys = data["keys"]
        for grp in self._data:
            keys = list(grp.keys())
            key = keys[0]
            if(len(keys) > 1 and keys[1] == 'fwv'):
                self._firmware_version = grp['fwv']
            if key not in DEVICE_TYPES:
                if key == "fwv":
                    self._firmware_version = grp[key]
                else:
                    _LOGGER.debug(f"Unknown device type: {key}")
                continue
            devices = grp[key]
            for device_data in devices:
                serial = device_data.get("sno")
                self._update_available = device_data.get('newBootloaderAvailable', False)
                existing_device = self.devices.get(serial, None)
                if existing_device is None:
                    existing_device = device_factory(
                        self._connection, key, serial, device_data
                    )
                    serial_key = existing_device.prefix + str(
                        existing_device.serial_number
                    )
                    existing_device.name = self.find_device_name(
                        serial_key,
                        f"{existing_device.kind}-{existing_device.serial_number}",
                    )
                    _LOGGER.debug(
                        f"Adding {existing_device.kind} {existing_device.name}"
                    )
                    self.devices[serial] = existing_device
                else:
                    _LOGGER.debug(
                        f"Updating {existing_device.kind} {existing_device.name}"
                    )
                    existing_device.data = device_data
        self._calculate_totals()

    async def refresh_history_today(self):
        today = datetime.now(timezone.utc)
        today = today.replace(hour=0, minute=0, second=0, microsecond=0)
        return await self.refresh_history(today, 24, HOUR)

    async def refresh_history(self, from_date, how_long, resolution):
        """Refresh history data for eddi and zappi"""
        devices = await self.get_devices("all", False)
        for device in devices:
            if device.kind == HARVI:
                continue
            await device.refresh_history_data(from_date, how_long, resolution)
        self._calculate_history_totals()

    async def fetch_data(self):
        """Fetch data from myenergi"""
        keys = self._keys
        if keys is None:
            keys = await self._connection.get("/cgi-get-app-key-")
        devices = await self._connection.get("/cgi-jstatus-*")
        data = {"devices": devices, "keys": keys}
        return data

    async def get_devices(self, kind="all", refresh=True):
        """Fetch devices, all or of a specific kind"""
        if refresh or not self.devices:
            await self.refresh()
        return self.get_devices_sync(kind)

    def get_devices_sync(self, kind="all"):
        """Return current devices"""
        all_devices = list(self.devices.values())
        if kind == "all":
            return all_devices
        return list(filter(lambda d: (d.kind == kind), all_devices))

    async def show(self):
        out = ""
        devices = await self.get_devices()
        await self.refresh_history_today()
        out = f"Site name: {self.site_name}\n"
        out = out + f"Home consumption             : {self.consumption_home}W\n"
        out = out + f"Power grid                   : {self.power_grid}W\n"
        out = out + f"Power generation             : {self.power_generation}W\n"
        out = out + f"Power EV charge              : {self.power_charging}W\n"
        out = out + f"Power battery                : {self.power_battery}W\n"
        out = out + f"Grid voltage                 : {self.voltage_grid}V\n"
        out = out + f"Grid frequency               : {self.frequency_grid}Hz\n"
        out = out + f"Energy imported              : {self.energy_imported}kWh\n"
        out = out + f"Energy exported              : {self.energy_exported}kWh\n"
        out = out + f"Energy generated             : {self.energy_generated}kWh\n"
        out = out + f"Energy green                 : {self.energy_green}kWh\n"
        out = out + "\n"
        out = out + f"Hub serial number            : {self.serial_number}\n"
        out = out + f"Hub firmware version         : {self.firmware_version}\n"
        out = out + "Devices:\n"
        for device in devices:
            out = out + f"\t{device.kind.capitalize()}: {device.name}"
            if device.kind != HARVI:
                out = out + f"\t{device.energy_total}kWh today\n"
                for key in device.ct_keys:
                    out = out + f"\t{key} {device.history_data.get(key, 0)}kWh today\n"
                out = out + "\n"
            else:
                out = out + "\n\n"
        return out



EDDI = "eddi"
ZAPPI = "zappi"
HARVI = "harvi"
LIBBI = "libbi"
DEVICE_TYPES = [EDDI, ZAPPI, HARVI, LIBBI]
HOUR = "hour"
MINUTE = "minute"
CT_GENERATION = "Generation"
CT_BATTERY = "AC Battery"
CT_LOAD = "Internal Load"
CT_MONITOR = "Monitor"
CT_GRID = "Grid"
FREQUENCY_GRID = "frequency"
VOLTAGE_GRID = "voltage"



from pymyenergi.connection import Connection

from . import HARVI
from .base_device import BaseDevice


class Harvi(BaseDevice):
    """Zappi Client for myenergi API."""

    def __init__(self, connection: Connection, serialno, data={}) -> None:
        super().__init__(connection, serialno, data)

    @property
    def kind(self):
        return HARVI

    @property
    def ct_keys(self):
        """Return CT key names that are not none"""
        keys = {}
        for i in range(3):
            ct = getattr(self, f"ct{i+1}")
            if ct.name_as_key == "ct_none":
                continue
            keys[ct.name_as_key] = keys.get(ct.name_as_key, 0) + 1
        return keys

    @property
    def ct3(self):
        """Current transformer 3"""
        return self._create_ct(3)

    @property
    def prefix(self):
        return "H"

    def show(self, short_format=False):
        """Returns a string with all data in human readable format"""
        ret = ""
        name = ""
        if self.name:
            name = f" {self.name}"
        ret = ret + f"Harvi S/N {self.serial_number}"
        ret = ret + f"{name}"
        if short_format:
            return ret
        ret = ret.center(80, "-") + "\n"
        ret = ret + f"CT 1 {self.ct1.name} {self.ct1.power}W phase {self.ct1.phase}\n"
        ret = ret + f"CT 2 {self.ct2.name} {self.ct2.power}W phase {self.ct2.phase}\n"
        ret = ret + f"CT 3 {self.ct3.name} {self.ct3.power}W phase {self.ct3.phase}\n"
        return ret



from pymyenergi.connection import Connection

from . import ZAPPI
from .base_device import BaseDevice

CHARGE_MODES = ["None", "Fast", "Eco", "Eco+", "Stopped"]
STATES = ["Unkn0", "Paused", "Unkn2", "Charging", "Boosting", "Completed"]
PLUG_STATES = {
    "A": "EV Disconnected",
    "B1": "EV Connected",
    "B2": "Waiting for EV",
    "C1": "EV ready to charge",
    "C2": "Charging",
    "F": "Fault",
    "U": "",
}


class Zappi(BaseDevice):
    """Zappi Client for myenergi API."""

    def __init__(self, connection: Connection, serialno, data=None) -> None:
        self.history_data = {}
        self.boost_data = {}
        super().__init__(connection, serialno, data)

    async def refresh(self):
        """Refresh device data"""
        self.data = await self.fetch_data()

    async def fetch_boost_data(self):
        """Fetch data from myenergi"""
        response = await self._connection.get(
            f"/cgi-boost-time-{self.prefix}{self._serialno}"
        )
        data = response
        return data

    @property
    def kind(self):
        return ZAPPI

    @property
    def prefix(self):
        return "Z"

    @property
    def ct_keys(self):
        """Return CT key names that are not none"""
        keys = {}
        for i in range(6):
            ct = getattr(self, f"ct{i+1}")
            if ct.name_as_key == "ct_none":
                continue
            keys[ct.name_as_key] = keys.get(ct.name_as_key, 0) + 1
        return keys

    @property
    def charge_mode(self):
        """Charge mode, one of Fast, Eco, Eco+ and Stopped"""
        return CHARGE_MODES[self._data.get("zmo", 0)]

    @property
    def charge_added(self):
        """Charge added this session in kWh"""
        return self._data.get("che")

    @property
    def is_dst(self):
        """Is DST in use"""
        return self._data.get("dat") == 1

    @property
    def ct3(self):
        """Current transformer 3"""
        return self._create_ct(3)

    @property
    def ct4(self):
        """Current transformer 4"""
        return self._create_ct(4)

    @property
    def ct5(self):
        """Current transformer 5"""
        return self._create_ct(5)

    @property
    def ct6(self):
        """Current transformer 6"""
        return self._create_ct(6)

    @property
    def supply_frequency(self):
        """Supply frequency in Hz"""
        return self._data.get("frq")

    @property
    def supply_voltage(self):
        """Supply voltage in V"""
        return self._data.get("vol", 0) / 10

    @property
    def power_grid(self):
        """Grid power in W"""
        return self._data.get("grd", 0)

    @property
    def power_generated(self):
        """Generated power in W"""
        return self._data.get("gen", 0)

    @property
    def status(self):
        """Current status, one of Paused, Charging or Completed"""
        return STATES[self._data.get("sta", 1)]

    @property
    def plug_status(self):
        """Plug status, one of EV Disconnected, EV Connected, Waiting for EV, EV Ready to charge, Charging or Fault"""
        return PLUG_STATES.get(self._data.get("pst", "U"), "")

    @property
    def priority(self):
        """Charger priority"""
        return self._data.get("pri", 0)

    @property
    def l1_phase(self):
        """What phase L1 is connected to"""
        return self._data.get("pha", 0)

    @property
    def locked(self):
        """Lock status"""
        return self._data.get("lck", 0) >> 1 & 1 == 1

    @property
    def lock_when_pluggedin(self):
        """Lock when plugged in status"""
        return self._data.get("lck", 0) >> 2 & 1 == 1

    @property
    def lock_when_unplugged(self):
        """Lock when unplugged status"""
        return self._data.get("lck", 0) >> 3 & 1 == 1

    @property
    def charge_when_locked(self):
        """Charge when locked enabled"""
        return self._data.get("lck", 0) >> 4 & 1 == 1

    @property
    def charge_session_allowed(self):
        """Allow charge override"""
        return self._data.get("lck", 0) >> 5 & 1 == 1

    @property
    def minimum_green_level(self):
        """Minimum green level"""
        return self._data.get("mgl", -1)

    @property
    def smart_boost_start_hour(self):
        """Smart boost starting at hour"""
        return self._data.get("sbh", -1)

    @property
    def smart_boost_start_minute(self):
        """Smart boost starting at minute"""
        return self._data.get("sbm", -1)

    @property
    def smart_boost_amount(self):
        """Smart boost amount of energy to add"""
        return self._data.get("sbk", -1)

    @property
    def energy_boost(self):
        """Device boosted energy from history data"""
        return self.history_data.get("boost_total", 0)
    
    @property
    def energy_green(self):
        """Device green energy from history data"""
        return self.history_data.get("device_green", 0)

    @property
    def energy_total(self):
        """Device total energy from history data"""
        return self.history_data.get("device_total", 0)

    # @property
    # def boost_start_hour(self):
    #     """Boost starting at hour ??"""
    #     return self._data.get("tbh", -1)

    # @property
    # def boost_start_minute(self):
    #     """Boost starting at minute ??"""
    #     return self._data.get("tbm", -1)

    @property
    def boost_amount(self):
        """Boost amount of energy to add"""
        return self._data.get("tbk", -1)

    # The following propterties are have unknown purpose, names will change once known
    @property
    def bst(self):
        return self._data.get("bst")

    @property
    def bsm(self):
        """Boost mode maybe, turns 1 when manual boosting"""
        return self._data.get("bsm")

    @property
    def bss(self):
        return self._data.get("bss")

    @property
    def tz(self):
        return self._data.get("tz")

    @property
    def pwm(self):
        return self._data.get("pwm", 0) / 100

    @property
    def zs(self):
        return self._data.get("zs")

    @property
    def rac(self):
        return self._data.get("rac")

    @property
    def rrac(self):
        return self._data.get("rrac")

    @property
    def zsh(self):
        return self._data.get("zsh")

    @property
    def zsl(self):
        return self._data.get("zsl")

    @property
    def update_available(self):
        return self._data.get("newBootloaderAvailable", False)

    @property
    def rdc(self):
        return self._data.get("rdc")

    def show(self, short_format=False):
        """Returns a string with all data in human readable format"""
        name = ""
        ret = ""
        if self.name:
            name = f" {self.name}"
        ret = ret + f"Zappi S/N {self.serial_number}"
        ret = ret + f"{name} version {self.firmware_version}"
        if self.update_available:
            ret = ret + " (update available)"
        if short_format:
            return ret
        ret = ret.center(80, "-") + "\n"
        ret = ret + f"Status: {self.status}\n"
        ret = ret + f"Plug status: {self.plug_status}\n"
        ret = ret + f"Locked: {self.locked}\n"
        ret = ret + f"Charge added: {self.charge_added}\n"
        ret = ret + f"Priority: {self.priority}\n"
        ret = ret + f"Charge mode: {self.charge_mode}\n"
        ret = ret + f"PWM: {self.pwm}%\n"
        ret = ret + "\n"
        ret = ret + f"Lock when plugged in   : {self.lock_when_pluggedin}\n"
        ret = ret + f"Lock when unplugged    : {self.lock_when_unplugged}\n"
        ret = ret + f"Charge when locked     : {self.charge_when_locked}\n"
        ret = ret + f"Charge session allowed : {self.charge_session_allowed}\n"
        ret = ret + "\n"
        ret = ret + f"CT 1 {self.ct1.name} {self.ct1.power}W\n"
        ret = ret + f"CT 2 {self.ct2.name} {self.ct2.power}W\n"
        ret = ret + f"CT 3 {self.ct3.name} {self.ct3.power}W\n"
        if self.ct4.name != "None":
            ret = ret + f"CT 4 {self.ct4.name} {self.ct4.power}W\n"
        if self.ct5.name != "None":
            ret = ret + f"CT 5 {self.ct5.name} {self.ct5.power}W\n"
        if self.ct6.name != "None":
            ret = ret + f"CT 6 {self.ct6.name} {self.ct6.power}W\n"
        ret = ret + "\n"
        ret = ret + f"Supply voltage: {self.supply_voltage}V\n"
        ret = ret + f"Line frequency: {self.supply_frequency}Hz\n"
        ret = ret + f"L1 phase: {self.l1_phase}\n"
        ret = ret + "Power:\n"
        ret = ret + f"  Grid      : {self.power_grid}W\n"
        ret = ret + f"  Generated : {self.power_generated}W\n"
        ret = ret + "\n"
        for key in self.ct_keys:
            ret = ret + f"Energy {key} {self.history_data.get(key, 0)}Wh\n"
        ret = ret + "\n"
        ret = ret + f"Boost with {self.boost_amount}kWh\n"
        ret = ret + "Smart Boost start at"
        ret = ret + f" {self.smart_boost_start_hour}:{self.smart_boost_start_minute}"
        ret = ret + f" add {self.smart_boost_amount}kWh\n"
        ret = ret + f"Minimum green level: {self.minimum_green_level}%"
        return ret

    async def stop_charge(self):
        """Stop charge"""
        await self._connection.get(f"/cgi-zappi-mode-Z{self._serialno}-4-0-0-0000")
        return True

    async def stop_boost(self):
        """Stop charge"""
        await self._connection.get(f"/cgi-zappi-mode-Z{self._serialno}-2-0-0-0000")
        return True

    async def set_charge_mode(self, mode):
        """Set charge mode, one of Fast, Eco, Eco+ or Stopped"""
        mode_int = CHARGE_MODES.index(mode.capitalize())
        await self._connection.get(
            f"/cgi-zappi-mode-Z{self._serialno}-{mode_int}-0-0-0000"
        )
        # Set local data if successful
        self._data["zmo"] = mode_int
        return True

    async def set_minimum_green_level(self, level):
        """Set minimum green level 0-100"""
        await self._connection.get(f"/cgi-set-min-green-Z{self._serialno}-{level}")
        # Set local data if successful
        self._data["mgl"] = level
        return True

    async def start_boost(self, amount):
        """Start boost"""
        if self.charge_mode not in ["Eco", "Eco+"]:
            return False
        await self._connection.get(
            f"/cgi-zappi-mode-Z{self._serialno}-0-10-{int(amount)}-0000"
        )
        return True

    async def set_priority(self, priority):
        """Set device priority"""
        await self._connection.get(
            f"/cgi-set-priority-Z{self._serialno}-{int(priority)}"
        )
        self._data["pri"] = int(priority)
        return True

    async def start_smart_boost(self, amount, complete_by):
        """Start smart boost"""
        time = complete_by.replace(":", "")
        await self._connection.get(
            f"/cgi-zappi-mode-Z{self._serialno}-0-11-{int(amount)}-{time}"
        )
        return True

    async def unlock(self):
        """Unlock"""
        await self._connection.get(f"/cgi-jlock-{self._serialno}-00000010")
        return True



import pytest
from pymyenergi.zappi import Zappi

pytestmark = pytest.mark.asyncio


async def test_refresh(zappi_fetch_data_fixture):
    """Test Zappi data"""
    zappi = Zappi({}, 16042300)
    await zappi.refresh()
    assert zappi.serial_number == 16042300
    assert zappi.charge_mode == "Fast"
    assert zappi.charge_added == 4.2



import logging

from pymyenergi.connection import Connection

from . import EDDI
from .base_device import BaseDevice

_LOGGER = logging.getLogger(__name__)

MODE_NORMAL = 1
MODE_STOPPED = 0
STATES = [
    "Unkn0",
    "Paused",
    "Unkn2",
    "Diverting",
    "Boosting",
    "Max temp reached",
    "Stopped",
]
BOOST_TARGETS = {"heater1": 1, "heater2": 2, "relay1": 11, "relay2": 12}
EDDI_MODES = ["Stopped", "Normal"]


class Eddi(BaseDevice):
    """Eddi Client for myenergi API."""

    def __init__(self, connection: Connection, serialno, data={}) -> None:
        self.history_data = {}
        super().__init__(connection, serialno, data)

    @property
    def kind(self):
        return EDDI

    @property
    def prefix(self):
        return "E"

    @property
    def heater_priority(self):
        """Current heater priority"""
        return self._data.get("hpri", 1)

    @property
    def ct_keys(self):
        """Return CT key names that are not none"""
        keys = {}
        for i in range(3):
            ct = getattr(self, f"ct{i+1}")
            if ct.name_as_key == "ct_none":
                continue
            keys[ct.name_as_key] = keys.get(ct.name_as_key, 0) + 1
        return keys

    @property
    def hsk(self):
        """ Heatsink temperature """
        val = self._data.get("hsk", None)
        if val is not None:
            val = val / 10
        return val

    @property
    def l1_phase(self):
        """What phase L1 is connected to"""
        return self._data.get("pha", 0)

    @property
    def status(self):
        """Current status, one of Paused, Charging or Completed"""
        return STATES[self._data.get("sta", 1)]

    @property
    def supply_frequency(self):
        """Supply frequency in Hz"""
        return self._data.get("frq")

    @property
    def supply_voltage(self):
        """Supply voltage in V"""
        return self._data.get("vol", 0) / 10

    @property
    def consumed_session(self):
        """Energy diverted this session kWh"""
        return self._data.get("che", 0)

    @property
    def power_grid(self):
        """Grid power in W"""
        return self._data.get("grd", 0)

    @property
    def power_generated(self):
        """Generated power in W"""
        return self._data.get("gen", 0)

    @property
    def energy_total(self):
        """Device total energy from history data"""
        return self.history_data.get("device_total", 0)

    @property
    def energy_green(self):
        """Device green energy from history data"""
        return self.history_data.get("device_green", 0)

    @property
    def temp_1(self):
        """Temperature probe 1 temp"""
        return self._data.get("tp1", -1)

    @property
    def temp_2(self):
        """Temperature probe 2 temp"""
        return self._data.get("tp2", -1)

    @property
    def temp_name_1(self):
        """Temperature probe 2 name"""
        return self._data.get("ht1", None)

    @property
    def temp_name_2(self):
        """Temperature probe 2 name"""
        return self._data.get("ht2", None)

    @property
    def priority(self):
        """Current priority"""
        return self._data.get("pri")

    @property
    def active_heater(self):
        """Active heater"""
        return self._data.get("hno")

    @property
    def remaining_boost_time(self):
        """For how much longer boost will be active in seconds"""
        return self._data.get("rbt", 0)

    @property
    def is_boosting(self):
        """For how much longer boost will be active in seconds"""
        return self._data.get("bsm", 0) == 1

    # CT1 and CT2 are defined in base device
    @property
    def ct3(self):
        """Current transformer 3"""
        return self._create_ct(3)

    # The following properties are unknonw, names might change
    @property
    def r1a(self):
        """r1a?"""
        return self._data.get("r1a")

    @property
    def r2a(self):
        """r2a?"""
        return self._data.get("r2a")

    @property
    def r1b(self):
        """r1b?"""
        return self._data.get("r1b")

    async def set_operating_mode(self, mode: str):
        """Stopped or normal mode"""
        mode_int = EDDI_MODES.index(mode.capitalize())
        await self._connection.get(f"/cgi-eddi-mode-E{self._serialno}-{mode_int}")
        if mode_int == 0:
            self._data["sta"] = 6
        else:
            self._data["sta"] = 5
        return True

    async def manual_boost(self, target: str, time: int):
        """Start manual boost of target for time minutes"""
        target_int = BOOST_TARGETS[target.lower().replace(" ", "")]
        await self._connection.get(
            f"/cgi-eddi-boost-E{self._serialno}-10-{target_int}-{time}"
        )
        return True

    async def set_priority(self, priority):
        """Set device priority"""
        await self._connection.get(
            f"/cgi-set-priority-E{self._serialno}-{int(priority)}"
        )
        self._data["pri"] = int(priority)
        return True

    async def set_heater_priority(self, target: str):
        """Start manual boost of target for time minutes"""
        target_int = BOOST_TARGETS[target.lower().replace(" ", "")]
        response = await self._connection.get(
            f"/cgi-set-heater-priority-E{self._serialno}"
        )
        cpm = response.get("cpm", 0)
        await self._connection.get(
            f"/cgi-set-heater-priority-E{self._serialno}-{target_int}-{cpm}"
        )
        self._data["hpri"] = target_int
        return True

    def show(self, short_format=False):
        """Returns a string with all data in human readable format"""
        ret = ""
        name = ""
        if self.name:
            name = f" {self.name}"
        ret = ret + f"Eddi S/N {self.serial_number}"
        ret = ret + f"{name} version {self.firmware_version}\n\n"
        if short_format:
            return ret
        ret = ret.center(80, "-") + "\n"
        ret = ret + f"Active heater:   {self.active_heater}\n"
        ret = ret + f"Eddi priority:   {self.priority}\n"
        ret = ret + f"Heater priority: {self.heater_priority}\n"
        if self.is_boosting:
            ret = ret + f"Boosting, {self.remaining_boost_time} miuntes left\n"
        ret = ret + f"CT 1 {self.ct1.name} {self.ct1.power}W\n"
        ret = ret + f"CT 2 {self.ct2.name} {self.ct2.power}W\n"
        ret = ret + f"CT 3 {self.ct3.name} {self.ct3.power}W\n"
        if self.hsk is not None:
            ret = ret + f"Heatsink temperature: {self.hsk}C\n"
        if self.temp_1 != -1:
            ret = ret + f"Temp {self.temp_name_1}: {self.temp_1}C\n"
        if self.temp_2 != -1:
            ret = ret + f"Temp {self.temp_name_2}: {self.temp_2}C\n"
        for key in self.ct_keys:
            ret = ret + f"Energy {key} {self.history_data.get(key, 0)}Wh\n"
        return ret



from unittest.mock import AsyncMock

import pytest
from pymyenergi.eddi import Eddi

pytestmark = pytest.mark.asyncio


async def test_refresh(eddi_fetch_data_fixture):
    """Test Zappi data"""
    eddi = Eddi({}, 16042300)
    await eddi.refresh()
    assert eddi.serial_number == 10088800
    assert eddi.consumed_session == 8.2


async def test_boost(eddi_fetch_data_fixture):
    """Test Zappi data"""
    eddi = Eddi(type("", (), {})(), 16042300)
    mock_get = AsyncMock()
    eddi._connection.get = mock_get
    await eddi.manual_boost("Relay 1", 400)
    mock_get.assert_awaited_with("/cgi-eddi-boost-E16042300-10-11-400")
    await eddi.manual_boost("Heater 1", 300)
    mock_get.assert_awaited_with("/cgi-eddi-boost-E16042300-10-1-300")
    await eddi.manual_boost("heater1", 300)
    mock_get.assert_awaited_with("/cgi-eddi-boost-E16042300-10-1-300")



#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for controlling Porsche Connect API.

"""
import logging

_LOGGER = logging.getLogger(__name__)


class MyenergiException(Exception):
    """Class of Porsche API exceptions."""

    def __init__(self, code=None, *args, **kwargs):
        """Initialize exceptions for the Porsche API."""
        self.message = ""
        super().__init__(*args, **kwargs)
        if code is not None:
            self.code = code
            if isinstance(code, str):
                self.message = self.code
                return
            if self.code == 401:
                self.message = "UNAUTHORIZED"
            elif self.code == 404:
                self.message = "NOT_FOUND"
            elif self.code == 405:
                self.message = "MOBILE_ACCESS_DISABLED"
            elif self.code == 408:
                self.message = "VEHICLE_UNAVAILABLE"
            elif self.code == 423:
                self.message = "ACCOUNT_LOCKED"
            elif self.code == 429:
                self.message = "TOO_MANY_REQUESTS"
            elif self.code == 500:
                self.message = "SERVER_ERROR"
            elif self.code == 503:
                self.message = "SERVICE_MAINTENANCE"
            elif self.code == 504:
                self.message = "UPSTREAM_TIMEOUT"
            elif self.code > 299:
                self.message = f"UNKNOWN_ERROR_{self.code}"


class WrongCredentials(MyenergiException):
    """Class of exceptions for incomplete credentials."""

    pass


class TimeoutException(MyenergiException):
    """Class of exceptions for incomplete credentials."""

    pass



import pytest
from pymyenergi.libbi import Libbi

pytestmark = pytest.mark.asyncio


async def test_refresh(libbi_fetch_data_fixture):
    """Test Libbi data"""
    libbi = Libbi({}, 24047164)
    await libbi.refresh()
    assert libbi.serial_number == 24047164



#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for connecting to myenergi API.

"""
import logging
import sys
from typing import Text

import httpx

from .exceptions import MyenergiException
from .exceptions import TimeoutException
from .exceptions import WrongCredentials

_LOGGER = logging.getLogger(__name__)


class Connection:
    """Connection to myenergi API."""

    def __init__(
        self, username: Text = None, password: Text = None, timeout: int = 20
    ) -> None:
        """Initialize connection object."""
        self.timeout = timeout
        self.director_url = "https://director.myenergi.net"
        self.base_url = None
        self.username = username
        self.password = password
        self.auth = httpx.DigestAuth(self.username, self.password)
        self.headers = {"User-Agent": "Wget/1.14 (linux-gnu)"}
        self.do_query_asn = True
        _LOGGER.debug("New connection created")

    def _checkMyenergiServerURL(self, responseHeader):
        if "X_MYENERGI-asn" in responseHeader:
            new_url = "https://" + responseHeader["X_MYENERGI-asn"]
            if new_url != self.base_url:
                _LOGGER.info(f"Updated myenergi active server to {new_url}")
            self.base_url = new_url
        else:
            _LOGGER.debug(
                "Myenergi ASN not found in Myenergi header, assume auth failure (bad username)"
            )
            raise WrongCredentials()

    async def send(self, method, url, json=None):
        # If base URL has not been set, make a request to director to fetch it

        async with httpx.AsyncClient(
            auth=self.auth, headers=self.headers, timeout=self.timeout
        ) as httpclient:
            if self.base_url is None or self.do_query_asn:
                _LOGGER.debug("Get Myenergi base url from director")
                try:
                    directorUrl = self.director_url + "/cgi-jstatus-E"
                    response = await httpclient.get(directorUrl)
                except Exception:
                    _LOGGER.error("Myenergi server request problem")
                    _LOGGER.debug(sys.exc_info()[0])
                else:
                    self.do_query_asn = False
                    self._checkMyenergiServerURL(response.headers)
            theUrl = self.base_url + url
            try:
                _LOGGER.debug(f"{method} {url} {theUrl}")
                response = await httpclient.request(method, theUrl, json=json)
            except httpx.ReadTimeout:
                # Make sure to query for ASN next request, might be a server problem
                self.do_query_asn = True
                raise TimeoutException()
            else:
                _LOGGER.debug(f"GET status {response.status_code}")
                self._checkMyenergiServerURL(response.headers)
                if response.status_code == 200:
                    return response.json()
                elif response.status_code == 401:
                    raise WrongCredentials()
                # Make sure to query for ASN next request, might be a server problem
                self.do_query_asn = True
                raise MyenergiException(response.status_code)

    async def get(self, url):
        return await self.send("GET", url)

    async def post(self, url, data=None):
        return await self.send("POST", url, data)

    async def put(self, url, data=None):
        return await self.send("PUT", url, data)

    async def delete(self, url, data=None):
        return await self.send("DELETE", url, data)



class CT:
    def __init__(self, name, value) -> None:
        self._name = name
        self._value = value

    @property
    def name(self):
        """Name of CT clamp"""
        return self._name

    @property
    def power(self):
        """Power reading of CT clamp in W"""
        return self._value



import pytest
from pymyenergi.client import MyenergiClient
from pymyenergi.eddi import Eddi
from pymyenergi.harvi import Harvi
from pymyenergi.zappi import Zappi
from pymyenergi.libbi import Libbi

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

conn = {}


async def test_init(bypass_client_fetch_data):
    client = MyenergiClient(conn)
    await client.refresh()
    assert len(client.devices) == 0


async def test_init_error(error_on_client_fetch_data):
    client = MyenergiClient(conn)
    with pytest.raises(Exception):
        assert await client.refresh()


async def test_get_all_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices()
    assert len(devices) == 6


async def test_get_eddi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("eddi")
    assert len(devices) == 1
    assert isinstance(devices[0], Eddi)


async def test_get_zappi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("zappi")
    assert len(devices) == 2
    assert isinstance(devices[1], Zappi)


async def test_get_harvi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("harvi")
    assert len(devices) == 2
    assert isinstance(devices[1], Harvi)


async def test_1p_harvi_eddi_solar_battery(client_1p_zappi_harvi_solar_battery_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("harvi")
    assert len(devices) == 1
    assert isinstance(devices[0], Harvi)
    devices = await client.get_devices("zappi")
    assert len(devices) == 1
    assert isinstance(devices[0], Zappi)
    assert client.power_grid == 10000
    assert client.power_generation == 5000
    assert client.power_battery == 3000
    assert client.power_charging == 2000
    assert client.consumption_home == 16000


async def test_get_libbi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("libbi")
    assert len(devices) == 1
    assert isinstance(devices[0], Libbi)


import logging
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from datetime import timedelta
from datetime import timezone

from pymyenergi.connection import Connection

from . import HOUR
from . import MINUTE

_LOGGER = logging.getLogger(__name__)


class CT:
    """Current Transformer class"""

    def __init__(self, name, value=0, phase=None) -> None:
        self._name = name
        self._value = value
        self._phase = phase

    @property
    def name(self):
        """Name of CT clamp"""
        return self._name

    @property
    def name_as_key(self):
        """Snake case version of name"""
        return "ct_" + self._name.replace(" ", "_").lower()

    @property
    def power(self):
        """Power reading of CT clamp in W"""
        return self._value

    @property
    def phase(self):
        """What phase the CT is assigned to"""
        return self._phase

    @property
    def is_assigned(self):
        """Is the CT assigned?"""
        return self.name != "None"

    @property
    def is_generation(self):
        return "generation" in self._name.lower()

    @property
    def is_grid(self):
        return "grid" in self._name.lower()


class BaseDevice(ABC):
    """Base class for myenergi devices"""

    def __init__(self, connection: Connection, serialno, data=None) -> None:
        self._connection = connection
        self._serialno = serialno
        self._data = data or {}
        self._name = None
        self.ct_groups = {}
        self.refresh_ct_groups()

    @property
    @abstractmethod
    def kind(self):
        """What kind of device"""

    @property
    @abstractmethod
    def prefix(self):
        """Device prefix used in api calls"""

    def _create_ct(self, ct_number):
        """Create a CT from data"""
        return CT(
            self._data.get(f"ectt{ct_number}", "None"),
            self._data.get(f"ectp{ct_number}", 0),
            self._data.get(f"ect{ct_number}p", None),
        )

    async def fetch_data(self):
        """Fetch data from myenergi"""
        response = await self._connection.get(
            f"/cgi-jstatus-{self.prefix}{self._serialno}"
        )
        data = response[self.kind][0]
        return data

    async def energy_today(self, raw_response=False):
        today = datetime.now(timezone.utc)
        today = today.replace(hour=0, minute=0, second=0, microsecond=0)
        # return await self.history_energy_minutes(today, 1440)
        return await self.history_energy_hours(today, 24, raw_response)

    async def history_energy_minutes(
        self, date_from, how_long=1440, raw_response=False
    ):
        if date_from is None:
            date_from = datetime.now(timezone.utc) - timedelta(minutes=how_long)
        return await self.fetch_history_data(date_from, how_long, MINUTE, raw_response)

    async def history_energy_hours(self, date_from, how_long=24, raw_response=False):
        if date_from is None:
            date_from = datetime.now(timezone.utc) - timedelta(hours=how_long)
        return await self.fetch_history_data(date_from, how_long, HOUR, raw_response)

    async def fetch_history_data(
        self, date_from, how_long, resolution, raw_response=False
    ):
        energy_wh = {
            "gep": 0,
            "gen": 0,
            "imp": 0,
            "exp": 0,
            "h1d": 0,
            "h1b": 0,
            "h2d": 0,
            "h2b": 0,
            "h3d": 0,
            "h3b": 0,
            "ct1": 0,
            "ct2": 0,
            "ct3": 0,
            "ct4": 0,
            "ct5": 0,
            "ct6": 0,
            "ive1": 0,
            "ivi1": 0,
            "bdp1": 0,
            "bcp1": 0,
            "pvp1": 0
        }
        if resolution == MINUTE:
            url = f"/cgi-jday-{self.prefix}{self._serialno}-{date_from.year}-{date_from.month}-{date_from.day}-{date_from.hour}-0-{how_long}"
        else:
            url = f"/cgi-jdayhour-{self.prefix}{self._serialno}-{date_from.year}-{date_from.month}-{date_from.day}-{date_from.hour}-{how_long}"
        _LOGGER.debug(f"Fetching {resolution} history data for {self.kind}")
        data = await self._connection.get(url)
        data = data[f"U{self.serial_number}"]
        if raw_response:
            return data

        for row in data:
            for key in energy_wh:
                if key in ["ct1", "ct2", "ct3", "ct4", "ct5", "ct6"]:
                    watt_hours = (
                        row.get(f"pe{key}", 0) / 3600 - row.get(f"ne{key}", 0) / 3600
                    )
                else:
                    watt_hours = row.get(key, 0) / 3600
                energy_wh[key] = energy_wh[key] + watt_hours

        device_boosted = round(
            (energy_wh["h1b"] + energy_wh["h2b"] + energy_wh["h3b"]) / 1000, 2
        )
        device_green = round(
            (energy_wh["h1d"] + energy_wh["h2d"] + energy_wh["h3d"]) / 1000, 2
        )

        return_data = {
            "generated": round(energy_wh["gep"] / 1000, 2),
            "grid_import": round(energy_wh["imp"] / 1000, 2),
            "grid_export": round(energy_wh["exp"] / 1000, 2),
            "battery_charge": round(energy_wh["bcp1"] / 1000, 2),
            "battery_discharge": round(energy_wh["bdp1"] / 1000, 2),
            "inverter_export": round(energy_wh["ive1"] /1000, 2),
            "inverter_import": round(energy_wh["ivi1"] /1000, 2),
            "device_boosted": device_boosted,
            "device_green": device_green,
            "device_total": device_boosted + device_green
        }
        if resolution == MINUTE:
            return_data["pv_total"] = round(energy_wh["pvp1"] / 1000, 2)

        for i in range(6):
            key = f"ct{i+1}"
            if hasattr(self, key):
                ct_key = getattr(self, key).name_as_key
                if ct_key != "ct_none":
                    return_data[ct_key] = round(
                        (return_data.get(ct_key, 0) + (energy_wh[f"ct{i+1}"] / 1000)), 2
                    )
        return return_data

    @property
    def name(self):
        """Name of device"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def serial_number(self):
        """Serial Number"""
        return self._data.get("sno", None)

    @property
    def firmware_version(self):
        """Firmware version"""
        return self._data.get("fwv", None)

    @property
    def date(self):
        """Device date"""
        return self._data.get("dat")

    @property
    def time(self):
        """Device time"""
        return self._data.get("tim")

    @property
    def ct1(self):
        """Current transformer 1"""
        return self._create_ct(1)

    @property
    def ct2(self):
        """Current transformer 2"""
        return self._create_ct(2)

    @property
    def data(self):
        """All device data"""
        return self._data

    def refresh_ct_groups(self):
        groups = {}
        for i in range(6):
            key = f"ct{i+1}"
            if hasattr(self, key):
                ct = getattr(self, key)
                if ct.name_as_key != "ct_none":
                    groups[ct.name_as_key] = groups.get(ct.name_as_key, 0) + ct.power
        self.ct_groups = groups

    @data.setter
    def data(self, value):
        """Set all device data"""
        self._data = value
        self.refresh_ct_groups()

    async def refresh_history_data(self, from_date, how_long, resolution):
        """Refresh device history data"""
        self.history_data = await self.fetch_history_data(
            from_date, how_long, resolution
        )

    async def refresh(self):
        """Refresh device data"""
        self.data = await self.fetch_data()

    def __str__(self):
        return f"{self.kind} S/N: {self._serialno}"

    def __repr__(self):
        return f"{self.kind}-{self._serialno}"



import argparse
import asyncio
import configparser
import json
import logging
import os
import sys
from getpass import getpass

from pymyenergi.client import device_factory
from pymyenergi.client import MyenergiClient
from pymyenergi.connection import Connection
from pymyenergi.eddi import BOOST_TARGETS
from pymyenergi.eddi import EDDI_MODES
from pymyenergi.exceptions import WrongCredentials
from pymyenergi.zappi import CHARGE_MODES
from pymyenergi.libbi import LIBBI_MODES

from . import EDDI
from . import HARVI
from . import ZAPPI
from . import LIBBI

logging.basicConfig()
logging.root.setLevel(logging.WARNING)


async def main(args):
    username = args.username or input("Please enter your hub serial number: ")
    password = args.password or getpass()
    conn = Connection(username, password)
    if args.debug:
        logging.root.setLevel(logging.DEBUG)
    client = MyenergiClient(conn)
    try:
        if args.version:
            ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
            version_file = open(os.path.join(ROOT_DIR, "VERSION"))
            version = version_file.read().strip()
            print(version)
            sys.exit(0)
        if args.command == "list":
            devices = await client.get_devices(args.kind)
            for device in devices:
                if args.json:
                    print(json.dumps(device.data, indent=2))
                else:
                    print(device.show(True))
        elif args.command == "overview":
            out = await client.show()
            print(out)
        elif args.command in [ZAPPI, EDDI, HARVI, LIBBI]:
            if args.serial is None:
                devices = await client.get_devices(args.command)
            else:
                devices = [device_factory(conn, args.command, args.serial)]
                await devices[0].refresh()
            for device in devices:
                if args.action == "show":
                    if args.json:
                        print(json.dumps(device.data, indent=2))
                    else:
                        print(device.show())
                elif args.action == "energy":
                    data = await device.energy_today(args.json)
                    if args.json:
                        print(json.dumps(data, indent=2))
                    else:
                        for key in data.keys():
                            print(f"{key}: {data[key]}kWh")
                elif args.action == "stop" and args.command == ZAPPI:
                    await device.stop_charge()
                    print("Charging was stopped")
                elif args.action == "mode" and args.command == ZAPPI:
                    if (
                        len(args.arg) < 1
                        or args.arg[0].capitalize() not in CHARGE_MODES
                    ):
                        modes = ", ".join(CHARGE_MODES)
                        sys.exit(f"A mode must be specifed, one of {modes}")
                    await device.set_charge_mode(args.arg[0])
                    print(f"Charging was set to {args.arg[0].capitalize()}")
                elif args.action == "mode" and args.command == EDDI:
                    if len(args.arg) < 1 or args.arg[0].capitalize() not in EDDI_MODES:
                        modes = ", ".join(EDDI_MODES)
                        sys.exit(f"A mode must be specifed, one of {modes}")
                    await device.set_operating_mode(args.arg[0])
                    print(f"Operating mode was set to {args.arg[0].capitalize()}")
                elif args.action == "mode" and args.command == LIBBI:
                    if len(args.arg) < 1 or args.arg[0].capitalize() not in LIBBI_MODES:
                        modes = ", ".join(LIBBI_MODES)
                        sys.exit(f"A mode must be specifed, one of {modes}")
                    await device.set_operating_mode(args.arg[0])
                    print(f"Operating mode was set to {args.arg[0].capitalize()}")
                elif args.action == "mingreen" and args.command == ZAPPI:
                    if len(args.arg) < 1:
                        sys.exit("A minimum green level must be provided")
                    await device.set_minimum_green_level(args.arg[0])
                    print(f"Minimum green level was set to {args.arg[0]}")
                elif args.action == "boost" and args.command == ZAPPI:
                    if await device.start_boost(args.arg[0]):
                        print(f"Start boosting with {args.arg[0]}kWh")
                    else:
                        print("Could not start boost, charge mode must be Eco or Eco+")
                elif args.action == "boost" and args.command == EDDI:
                    if len(args.arg) < 2 or args.arg[0] not in BOOST_TARGETS:
                        targets = ", ".join(BOOST_TARGETS)
                        sys.exit(
                            f"A boost target and time must be specifed, one of {targets}"
                        )
                    if await device.manual_boost(args.arg[0], args.arg[1]):
                        print(f"Start boosting {args.arg[0]} for {args.arg[1]} minutes")
                    else:
                        print("Could not start boost")
                elif args.action == "priority" and args.command in [EDDI, ZAPPI, LIBBI]:
                    if len(args.arg) < 1:
                        sys.exit("A priority must be specifed, a number")
                    if await device.set_priority(args.arg[0]):
                        print(f"Device priority was set to {args.arg[0]}")
                    else:
                        print("Could not set heater priority")
                elif args.action == "heaterpriority" and args.command == EDDI:
                    if len(args.arg) < 1 or args.arg[0] not in BOOST_TARGETS:
                        targets = ", ".join(BOOST_TARGETS)
                        sys.exit(
                            f"A priority target must be specifed, one of {targets}"
                        )
                    if await device.set_heater_priority(args.arg[0]):
                        print(f"Heater priority was set to {args.arg[0]}")
                    else:
                        print("Could not set heater priority")
                elif args.action == "smart-boost" and args.command == ZAPPI:
                    if await device.start_smart_boost(args.arg[0], args.arg[1]):
                        print(
                            f"Start smart boosting with {args.arg[0]}kWh complete by {args.arg[1]}"
                        )
                    else:
                        print(
                            "Could not start smart boost, charge mode must be Eco or Eco+"
                        )
        else:
            sys.exit(
                "Dont know what to do, type myenergi --help form available commands"
            )
    except WrongCredentials:
        sys.exit("Wrong username or password")


def cli():
    config = configparser.ConfigParser()
    config["hub"] = {"serial": "", "password": ""}
    config.read([".myenergi.cfg", os.path.expanduser("~/.myenergi.cfg")])
    parser = argparse.ArgumentParser(prog="myenergi", description="myenergi CLI.")
    parser.add_argument(
        "-u",
        "--username",
        dest="username",
        default=config.get("hub", "serial"),
    )
    parser.add_argument(
        "-p",
        "--password",
        dest="password",
        default=config.get("hub", "password").strip('"'),
    )
    parser.add_argument("-d", "--debug", dest="debug", action="store_true")
    parser.add_argument("-j", "--json", dest="json", action="store_true", default=False)
    parser.add_argument("--version", dest="version", action="store_true", default=False)
    subparsers = parser.add_subparsers(dest="command", help="sub-command help")
    subparser_list = subparsers.add_parser("list", help="list devices")
    subparser_list.add_argument("-k", "--kind", dest="kind", default="all")
    subparsers.add_parser("overview", help="show overview")
    subparser_zappi = subparsers.add_parser(
        ZAPPI, help="use zappi --help for available commands"
    )
    subparser_zappi.add_argument("-s", "--serial", dest="serial", default=None)
    subparser_zappi.add_argument(
        "action",
        choices=[
            "show",
            "energy",
            "stop",
            "mode",
            "boost",
            "smart-boost",
            "mingreen",
            "priority",
            "unlock",
        ],
    )
    subparser_zappi.add_argument("arg", nargs="*")
    subparser_eddi = subparsers.add_parser(
        EDDI, help="use eddi --help for available commands"
    )
    subparser_eddi.add_argument("-s", "--serial", dest="serial", default=None)
    subparser_eddi.add_argument(
        "action",
        choices=["show", "energy", "mode", "boost", "heaterpriority", "priority"],
    )
    subparser_eddi.add_argument("arg", nargs="*")

    subparser_harvi = subparsers.add_parser(
        HARVI, help="use harvi --help for available commands"
    )
    subparser_harvi.add_argument("-s", "--serial", dest="serial", default=None)
    subparser_harvi.add_argument("action", choices=["show"])
    subparser_harvi.add_argument("arg", nargs="*")

    subparser_libbi = subparsers.add_parser(
        LIBBI, help="use libbi --help for available commands"
    )
    subparser_libbi.add_argument("-s", "--serial", dest="serial", default=None)
    subparser_libbi.add_argument("action", choices=["show","mode","priority","energy"])
    subparser_libbi.add_argument("arg", nargs="*")

    args = parser.parse_args()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(args))



#!/usr/bin/env python
import os

from setuptools import setup

version_file = open(os.path.join("pymyenergi", "VERSION"))
version = version_file.read().strip()

setup(
    name="pymyenergi",
    version=version,
    author="Johan Isaksson",
    author_email="johan@generatorhallen.se",
    description="Python library and CLI for communicating with myenergi API.",
    long_description=open("README.md", 'r').read(),
    long_description_content_type='text/markdown',
    package_data={"pymyenergi": ["VERSION"]},
    include_package_data=True,
    url="https://github.com/cjne/pymyenergi",
    license="MIT",
    packages=["pymyenergi"],
    python_requires=">=3.6",
    install_requires=["httpx"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Operating System :: OS Independent",
    ],
    setup_requires=("pytest-runner"),
    tests_require=(
        "asynctest",
        "pytest-cov",
        "pytest-asyncio",
        "pytest-trio",
        "pytest-tornasync",
    ),
)



import logging

from pymyenergi.connection import Connection

from . import LIBBI
from .base_device import BaseDevice

_LOGGER = logging.getLogger(__name__)

MODE_NORMAL = 1
MODE_STOPPED = 0

STATES = { 0:'Off',
           1:'On',
           2:'Battery Full',
           4:'Idle',
           5:'Charging',
           6:'Discharging',
           7:'Duration Charging',
           101:'Idle?',
           102:'102',
           234:'Calibration Charge' }

LIBBI_MODES = ["Stopped","Normal"]
LIBBI_MODE_NAMES = ["STOP", "BALANCE"]

class Libbi(BaseDevice):
    """Libbi Client for myenergi API."""

    def __init__(self, connection: Connection, serialno, data={}) -> None:
        self.history_data = {}
        super().__init__(connection, serialno, data)

    @property
    def kind(self):
        return LIBBI
        
    @property
    def status(self):
        """Get current known status"""
        n = self._data.get("sta", 1)
        if n in STATES:
            return STATES[n]
        else:
            return n
    
    @property
    def local_mode(self):
        """Get current known status"""
        return self._data.get("lmo", 1)

    @property
    def prefix(self):
        return "E"

    @property
    def ct_keys(self):
        """Return CT key names that are not none"""
        keys = {}
        for i in range(6):
            ct = getattr(self, f"ct{i+1}")
            if ct.name_as_key == "ct_none":
                continue
            keys[ct.name_as_key] = keys.get(ct.name_as_key, 0) + 1
        return keys

    @property
    def ct3(self):
        """Current transformer 3"""
        return self._create_ct(3)
    
    @property
    def ct4(self):
        """Current transformer 4"""
        return self._create_ct(4)
    
    @property
    def ct5(self):
        """Current transformer 4"""
        return self._create_ct(5)
    
    @property
    def ct6(self):
        """Current transformer 4"""
        return self._create_ct(6)
    
    @property
    def supply_frequency(self):
        """Supply frequency in Hz"""
        return self._data.get("frq")

    @property
    def supply_voltage(self):
        """Supply voltage in V"""
        return self._data.get("vol", 0) / 10

    @property
    def consumed_session(self):
        """Energy diverted this session kWh"""
        return self._data.get("che", 0)

    @property
    def power_grid(self):
        """Grid power in W"""
        return self._data.get("grd", 0)

    @property
    def power_generated(self):
        """Generated power in W"""
        return self._data.get("gen", 0)

    @property
    def energy_total(self):
        """Device total energy from history data"""
        return self.history_data.get("device_total", 0)

    @property
    def energy_green(self):
        """Device green energy from history data"""
        return self.history_data.get("device_green", 0)
    
    @property
    def state_of_charge(self):
        """State of Charge in %"""
        return self._data.get("soc", 0)
    
    @property
    def priority(self):
        """Current priority"""
        return self._data.get("pri")

    @property
    def battery_size(self):
        """Battery size in kwh"""
        return self._data.get("mbc", 0) /1000
    
    @property
    def inverter_size(self):
        """Inverter size in kwh"""
        return self._data.get("mic", 0) /1000

    @property
    def grid_import(self):
        """Grid import from history data"""
        return self.history_data.get("grid_import", 0)

    @property
    def grid_export(self):
        """Grid export from history data"""
        return self.history_data.get("grid_export", 0)

    @property
    def battery_charge(self):
        """Battery charge from history data"""
        return self.history_data.get("battery_charge", 0)

    @property
    def battery_discharge(self):
        """Battery discharge from history data"""
        return self.history_data.get("battery_discharge", 0)

    @property
    def generated(self):
        """Solar generation from history data"""
        return self.history_data.get("generated", 0)

    @property
    def prefix(self):
        return "L"
    

    async def set_operating_mode(self, mode: str):
        """Stopped or normal mode"""
        print("current mode", self._data["lmo"])
        mode_int = LIBBI_MODES.index(mode.capitalize())
        await self._connection.get(
            f"/cgi-libbi-mode-{self.prefix}{self._serialno}-{mode_int}"
            )
        self._data["lmo"] = LIBBI_MODE_NAMES[mode_int]
        return True
    
    async def set_priority(self, priority):
        """Set device priority"""
        await self._connection.get(
            f"/cgi-set-priority-{self.prefix}{self._serialno}-{int(priority)}"
        )
        self._data["pri"] = int(priority)
        return True

    def show(self, short_format=False):
        """Returns a string with all data in human readable format"""
        ret = ""
        name = ""
        if self.name:
            name = f" {self.name}"
        ret = ret + f"Libbi S/N {self.serial_number}"
        ret = ret + f"{name}"
        if short_format:
            return ret
        ret = ret.center(80, "-") + "\n"
        ret = ret + f"Libbi priority: {self.priority}\n"
        ret = ret + f"Battery size: {self.battery_size}kWh\n"
        ret = ret + f"Inverter size: {self.inverter_size}kWh\n"
        ret = ret + f"State of Charge: {self.state_of_charge}%\n"
        ret = ret + f"Generating: {self.power_generated}W\n" 
        ret = ret + f"Grid: {self.power_grid}W\n"
        ret = ret + f"Status : {self.status}\n"
        ret = ret + f"Local Mode : {self.local_mode}\n"
        ret = ret + f"CT 1 {self.ct1.name} {self.ct1.power}W phase {self.ct1.phase}\n"
        ret = ret + f"CT 2 {self.ct2.name} {self.ct2.power}W phase {self.ct2.phase}\n"
        ret = ret + f"CT 3 {self.ct3.name} {self.ct3.power}W phase {self.ct3.phase}\n"
        ret = ret + f"CT 4 {self.ct4.name} {self.ct4.power}W phase {self.ct4.phase}\n"
        ret = ret + f"CT 5 {self.ct5.name} {self.ct5.power}W phase {self.ct5.phase}\n"
        ret = ret + f"CT 6 {self.ct6.name} {self.ct6.power}W phase {self.ct6.phase}\n"
        for key in self.ct_keys:
            ret = ret + f"Energy {key} {self.history_data.get(key, 0)}Wh\n"
        return ret



import pytest
from pymyenergi.harvi import Harvi

pytestmark = pytest.mark.asyncio


async def test_refresh(harvi_fetch_data_fixture):
    """Test Harvi data"""
    harvi = Harvi({}, 10645200)
    await harvi.refresh()
    assert harvi.serial_number == 10645200
